// dllmain.cpp : Defines the entry point for the DLL application.
#include "pch.h"
#include "dllmain.h"
#include <string>
#include <vector>
#include <thread>
#include "osc/OscReceivedElements.h"
#include "osc/OscPacketListener.h"
#include "ip/UdpSocket.h"

//Class from OSC library (oscpack). This calls our handleOscMsg() on every new message.
class ThePacketListener : public osc::OscPacketListener {
protected:

    virtual void ProcessMessage(const osc::ReceivedMessage& m,
        const IpEndpointName& remoteEndpoint)
    {
        (void)remoteEndpoint; // suppress unused parameter warning

        try { 
            handleOscMsg(m); 
        }
        catch (osc::Exception& e) {
            // any parsing errors such as unexpected argument types, or 
            // missing arguments get thrown as exceptions.
            OutputDebugString(L"ProcessMessage: Error while parsing message: ...");
        }
    }
};

/* Definition and declaration of the registered OSC Listeners. 
* Contains OSC addresses to listen to and OS message ID to send for each 
* address. These are populated from the AHK script! */
struct Listener {
    std::string address;
    unsigned int dataType;
    unsigned int message;
};
std::vector<Listener> listeners;
/* Handle to the calling window, also indicates if this OSC receiver is "opened" by being != NULL*/
HWND hwnd = NULL;
/* OSC receiving objects and thread.
* From within these, our handleOscMsg() function is called. */
ThePacketListener* thePacketListener;
UdpListeningReceiveSocket* sock;
std::thread* oscThread;

/* The OSC Thread. Has to be another thread, because
*  UdpListeningReceiveSocket runs in a blocking loop. [RunUntilSigInt()]*/
void runOscThread(unsigned int port)
{
    thePacketListener = new ThePacketListener;
    sock = new UdpListeningReceiveSocket(
        IpEndpointName(IpEndpointName::ANY_ADDRESS, port),
        thePacketListener);
    sock->RunUntilSigInt(); //<<--- this is the loop

    //Cleanup after sock exited the loop
    sock = NULL;
    thePacketListener = NULL;
}

/* DLL was loaded */
BOOL APIENTRY DllMain( HMODULE hModule,
                       DWORD  ul_reason_for_call,
                       LPVOID lpReserved
                     )
{
    switch (ul_reason_for_call)
    {
    case DLL_PROCESS_ATTACH:
    case DLL_THREAD_ATTACH:
        break;
    case DLL_THREAD_DETACH:
    case DLL_PROCESS_DETACH:
        close();
        break;
    }
    return TRUE;
}

DLLEXPORT int open(HWND targetWindowHandle, unsigned int port)
{
    OutputDebugString(L"open()\r\n");
    if (!targetWindowHandle || !port) {
        if (!targetWindowHandle) OutputDebugString(L"open: targetWindowHandle!!\r\n");
        if (!port) OutputDebugString(L"open: port!!\r\n");
        return 1;
    }

    /*Store handle to Autohotkey window globally*/
    hwnd = targetWindowHandle;
    /*Start OSC Thread*/
    oscThread = new std::thread(runOscThread, port);

    OutputDebugString(L"open: Opened.\r\n");
    return 0;
}

DLLEXPORT int close()
{
    if (sock && sock->IsBound()) 
    {
        sock->AsynchronousBreak();
    }
    oscThread = NULL;
    hwnd = NULL;
    listeners.clear();
    return 0;
}

DLLEXPORT int addListener(LPCSTR address_, unsigned int messageID_, unsigned int dataType_)
{
    OutputDebugString(L"addListener: address=");
    OutputDebugStringA(address_);
    OutputDebugString(L"\r\n");
    listeners.push_back(Listener{ std::string(address_), dataType_, messageID_ });
    return 0;
}

DLLEXPORT int removeListener(LPCSTR address_)
{
    UINT result = 1;
    std::string addrStr(address_);
    //All OSC addresses have to start with a '/'
    if (addrStr[0] != '/') addrStr.insert(0, "/");

    for (UINT i = 0; i < listeners.size(); i++)
    {
        if (listeners[i].address == addrStr)
        {
            listeners.erase(listeners.begin() + i);
            i--; //one step backward, what previously was i+1 is i after erase().
            result = 0; //Found at least one occurence
        }
    }
    return result;
}

DLLEXPORT int handleOscMsg(const osc::ReceivedMessage& m)
{
    int ret = 0;

    OutputDebugString(L"handleOscMsg: Address=");
    OutputDebugStringA(m.AddressPattern());
    OutputDebugString(L"\r\n");

    if (!hwnd || !sock->IsBound())
    {
        OutputDebugString(L"handleOscMsg: Seems open() was not called yet...\r\n");
        return -1;
    }

    //Loop through all registered listener entries
    for (UINT i = 0; i < listeners.size(); i++)
    {
        //Check if incoming OSC address matches current listener entry
        if (listeners[i].address.compare(m.AddressPattern()) == 0)
        {
            //No payload
            if (m.ArgumentCount() == 0) {
                if (listeners[i].dataType == OSC_TYPE_NONE || listeners[i].dataType == OSC_TYPE_ALL) {
                    PostMessage(hwnd, listeners[i].message, OSC_TYPE_NONE, 0);
                    ret++;
                }
            }
            else //Some paylad, only check first one
            {
                osc::ReceivedMessage::const_iterator arg = m.ArgumentsBegin();
                if (arg->IsInt32()) { //Got Int, check if listener is valid for Int or All
                    if (listeners[i].dataType == OSC_TYPE_ALL || listeners[i].dataType == OSC_TYPE_INT) {
                        PostMessage(hwnd, listeners[i].message, OSC_TYPE_INT, arg->AsInt32()); //post to message queue
                        ret++;
                    }
                }
                else if (arg->IsFloat()) { //Got Float, check if listener is valid for Float or All
                    if (listeners[i].dataType == OSC_TYPE_ALL || listeners[i].dataType == OSC_TYPE_FLOAT) {
                        float floatarg = arg->AsFloat();
                        LPARAM lParam = *(LPARAM*)&floatarg;
                        PostMessage(hwnd, listeners[i].message, OSC_TYPE_FLOAT, lParam); //post to message queue
                        ret++;
                    }
                }
                else
                    OutputDebugString(L"handleOscMsg: Unknown datatype\r\n");
            }
        }
    }

    if (ret > 0) OutputDebugString(L"handleOscMsg: Found and posted at least once.\r\n");

    return ret;
}

/* Debugging function to test messaging. Probably will be removed later. */
DLLEXPORT int testMsg(HWND windowHandle, unsigned int messageID)
{
    OutputDebugString(L"testMsg()\r\n");
    if (!windowHandle) 
    {
        OutputDebugString(L"testMsg: windowHandle!!\r\n");
        return 1;
    }

    if (PostMessage(windowHandle, messageID, 42, 43))
        return 0;
    else
        return 1;
}

